//-----------------------------------------------------------------------------
//
// Author: John Clayton
// Date  : Nov.   5, 2001
// Update: Nov.   5, 2001 Obtained this file from "build_9b" project.
// Update: Nov.  30, 2001 First physical synthesis. (only "store" implemented)
// Update: Dec.  14, 2001 "Loads" and "stores" implemented.  Still debugging.
// Update: Dec.  18, 2001 Finished debugging.  "memory_sizer_dual_path" works
//                        at 25 MHz, requires 300 slices.  "memory_sizer"
//                        requires only 200 slices, but operates slower.
//                        (tested at 12.5 MHz only.  It worked fine.)
//
// Description
//-----------------------------------------------------------------------------
// This targets an XC2S200 board which was created for educational purposes.
//
// There are:
//    8  LEDs (led[7:0])
//    4  switches (switch[3:0])
//    1  clock of 32.000 MHz clock, present on GCLK1
//    1  clock of 49.152 MHz clock, present on GCLK0
//    4  lines of ps2 clock input (port A in documentation notes)
//    4  lines of ps2 data input (port A in documentation notes)
//    16 lines of LCD panel control (port B in documentation notes)
//    2  lines of rs232 serial connection (port C in documentation notes)
//-----------------------------------------------------------------------------

// NOTE: This build is for testing out a "memory_sizer" block.
//       The memory_sizer block allows a microprocessor to generate different
//       types of memory accesses (8-bit, 16-bit, 32-bit etc.) to different
//       and various sizes of memory.  It handles the "byte steering" and
//       multiple cycle action needed in order to accomplish these accesses
//       for the microprocessor.  In this build, the rs232_syscon acts as the
//       microprocessor (stand alone).
//
//       The dual-ported RAMs exhibit "little endian" transformation.
//
//       That is:  Sequential bytes, 00, 01, 02, 03 on the s8 side emerge
//                 on the s16 side:  01000302.
//                 on the s32 side:  03020100.
//
//                 Sequential words, 0100, 0302 on the s16 side emerge
//                 on the s32 side:  03020100.
//
//       This is called "little endian".  It can be quite confusing at times!

// The following 'include line must be used with Synplify to create EDIF
// The line must be commented for ModelSim.
//`include "d:\synplicity\synplify_70\lib\xilinx\virtex.v"

module top (
  sys_clk_0,
  sys_clk_1,
  switch,
  led,
  ps2_clk,
  ps2_data,
  lcd_drive,
  rs232_rxd,
  rs232_txd
  );


// I/O declarations
input sys_clk_0;      // 49.152 MHz
input sys_clk_1;      // 32.000 MHz
input [3:0] switch;
input rs232_rxd;

inout [3:0] ps2_clk;
inout [3:0] ps2_data;

output [7:0] led;

output [15:0] lcd_drive;
output rs232_txd;

// Internal signal declarations
wire [11:0] r0_wire;    // "Read" reg.  Used to "hold" pin locations, so that
                        // the synthesis tools do not complain about these pins
                        // being present in the constraints file and not in the
                        // design...

wire [9:0] crosshair_x_pos;
wire [9:0] crosshair_y_pos;

wire [7:0] lcd_control;

// Clock signals
wire clk_s;             // 49.152/2 MHz = 24.576 MHz
reg [1:0] clk_count;    // Counter used to generate clock

     // Signals from rs232_syscon
wire [31:0] adr;
wire [31:0] dat;
wire we;
wire stb;
wire rst;
wire master_br;

    // Memory sizer signals
wire [31:0] mem_adr;
wire [31:0] mem_dat;
wire [3:0] mem_be;
wire [1:0] access_width;  // 2:4 decoder is used at memory_sizer
wire [2:0] access_width_decode;
wire mem_we;
wire access_big_endian;
wire exc_be;
wire exc_wd;

    // Address decode signals
wire        selio;       // High for access to I/O space
wire [7:0]  sel;         // 1 of 8 is active high for I/O space accesses
wire [7:0]  ram_sel;     // 1 of 8 is active high for RAM space accesses
wire        ram_be_area; // RAM where memory sizer is used (byte enables)
wire        ram_a_area;  // RAM where memory sizer is not used (side "a")

    // RAM signals
wire [7:0] ram_0a_dat_o;
wire [15:0] ram_1a_dat_o;
wire [31:0] ram_2a_dat_o;
wire [7:0] ram_0b_dat_o;
wire [15:0] ram_1b_dat_o;
wire [31:0] ram_2b_dat_o;

    // Other...
wire reset = switch[0];  // Simply a renaming exercise

//--------------------------------------------------------------------------
// Clock generation
//--------------------------------------------------------------------------

// This uses up an additional GCLK resource.
always @(posedge sys_clk_0)
begin
  clk_count <= clk_count + 1;
end
assign clk_s = clk_count[0];

//--------------------------------------------------------------------------
// Instantiations
//--------------------------------------------------------------------------

lcd_test_pongball_crosshair lcd_block (
  .sys_clk(sys_clk_1),
  .lcd_clk(sys_clk_0),
  .sys_reset(lcd_control[1]),
  .lcd_reset(lcd_control[0]),
  .ball_x_enable(lcd_control[2]),
  .ball_y_enable(lcd_control[3] && switch[3]),  // switch[3] not needed!
  .crosshair_x_pos(crosshair_x_pos),
  .crosshair_y_pos(crosshair_y_pos),
  .crosshair_color(lcd_control[6:4]),
  .lcd_drive(lcd_drive)
  );


rs232_syscon #(
               8,             // Number of Hex digits for addresses.
               8,             // Number of Hex digits for data.
               2,             // Number of Hex digits for quantity.
               32,            // Characters in the input buffer
               5,             // Bits in the buffer pointer
               255,           // Clocks before watchdog timer expires
               8,             // Bits in watchdog timer
               4,             // Number of data fields displayed per line
               3,             // Number of bits in the fields counter
               3              // Number of bits in the digits counter
               )
  syscon_1 (                  // instance name
  .clk_i(clk_s),
  .reset_i(reset),
  .master_bg_i(master_br),
  .ack_i(selio || ram_a_area || ram_access_ack),
  .err_i(exc_be || exc_wd),
  .master_adr_i(32'b0),
  .master_stb_i(1'b0),
  .master_we_i(1'b0),
  .rs232_rxd_i(rs232_rxd),
  .dat_io(dat),
  .rst_o(rst),
  .master_br_o(master_br),
  .stb_o(stb),
  .cyc_o(),
  .adr_o(adr),
  .we_o(we),
  .rs232_txd_o(rs232_txd)
  );


assign access_width_decode = 1 << access_width;
memory_sizer_dual_path 
  memory_sizer_1 (
  .clk_i(clk_s),
  .reset_i(rst),
  .sel_i(ram_be_area),
  .memory_ack_i(| ram_sel[6:4]),
  .memory_has_be_i(1'b1),
  .memory_width_i(ram_sel[6:4]),  // The encoding is automatic!!!
  .access_width_i(access_width_decode),
  .access_big_endian_i(access_big_endian),
  .adr_i(adr),
  .we_i(we),
  .dat_io(dat),
  .memory_dat_io(mem_dat),
  .memory_adr_o(mem_adr),
  .memory_we_o(mem_we),
  .memory_be_o(mem_be),
  .access_ack_o(ram_access_ack),
  .exception_be_o(exc_be),
  .exception_watchdog_o(exc_wd)
  );


// I/O decodes
assign      selio   = adr[31:8] == 8'hFF && stb;
assign      sel[0]  = adr[7:5] == 0 && selio;
assign      sel[1]  = adr[7:5] == 1 && selio;
assign      sel[2]  = adr[7:5] == 2 && selio;
assign      sel[3]  = adr[7:5] == 3 && selio;
assign      sel[4]  = adr[7:5] == 4 && selio;
assign      sel[5]  = adr[7:5] == 5 && selio;
assign      sel[6]  = adr[7:5] == 6 && selio;
assign      sel[7]  = adr[7:5] == 7 && selio;

// RAM decodes

// The bits are as follows, from lsb to msb...
// Each line actually decodes 4096 bytes.  Values shown below are the actual
// sizes of the memory blocks connected to the decode lines...
// 0000 - 01ff hex (512 bytes  =  512 bytes) - direct "side A"
// 1000 - 13ff hex (512 words  = 1024 bytes) - direct "side A"
// 2000 - 27ff hex (512 dwords = 2048 bytes) - direct "side A"
// 3000 - 3fff hex (unused, space for 4096 bytes)
// 4000 - 41ff hex (512 bytes  =  512 bytes) - mem_sizer "side B"             
// 5000 - 53ff hex (512 words  = 1024 bytes) - mem_sizer "side B"
// 6000 - 67ff hex (512 dwords = 2048 bytes) - mem_sizer "side B"
// 7000 - 7fff hex (unused, space for 4096 bytes)

assign  ram_sel[0]  = ((adr[31:12] == 0) && stb);
assign  ram_sel[1]  = ((adr[31:12] == 1) && stb);
assign  ram_sel[2]  = ((adr[31:12] == 2) && stb);
assign  ram_sel[3]  = ((adr[31:12] == 3) && stb);
assign  ram_sel[4]  = ((adr[31:12] == 4) && stb);
assign  ram_sel[5]  = ((adr[31:12] == 5) && stb);
assign  ram_sel[6]  = ((adr[31:12] == 6) && stb);
assign  ram_sel[7]  = ((adr[31:12] == 7) && stb);
assign  ram_a_area = | ram_sel[2:0];
assign  ram_be_area = | ram_sel[7:4]; // Unused space included for testing
                                      // of mem_sizer watchdog...


assign r0_wire = {switch,~ps2_clk,~ps2_data};

reg_8_pack #(
             12,              // Size of r0
             2,               // Size of r1
             1,               // Size of r2
             1,               // Size of r3
             8,               // Size of r4
             10,              // Size of r5
             10,              // Size of r6
             8,               // Size of r7
             1,               // Read only regs.
             16               // Size of the data bus.
             )
  reg_8_pack_1                // Instance name
  (
   .clk_i(clk_s),
   .rst_i(reset),
   .sel_i(sel[0]),
   .we_i(we),
   .adr_i(adr[3:1]),  // byte addressed... word accessible only
   .dat_io(dat[15:0]),
   .r0(r0_wire),
   .r1(access_width),
   .r2(access_big_endian),
   .r3(),
   .r4(led),
   .r5(crosshair_x_pos),
   .r6(crosshair_y_pos),
   .r7(lcd_control)
   );


// 512 bytes of RAM (512 x 8)
assign dat[7:0] = (ram_sel[0] && ~we)?ram_0a_dat_o:{8{1'bZ}};
assign mem_dat[7:0] = (ram_sel[4] && ~mem_we)?ram_0b_dat_o:{8{1'bZ}};

RAMB4_S8_S8
   dp_ram_s8_1
   (
  .DOA(ram_0a_dat_o),
  .ADDRA(adr[8:0]),
  .CLKA(sys_clk_0),
  .DIA(dat[7:0]),
  .ENA(1'b1),
  .RSTA(reset),
  .WEA(we & ram_sel[0]),
  .DOB(ram_0b_dat_o),
  .ADDRB(mem_adr[8:0]),
  .CLKB(sys_clk_0),
  .DIB(mem_dat[7:0]),
  .ENB(mem_be[0]),
  .RSTB(reset),
  .WEB(mem_we & ram_sel[4])
  );


// 1k bytes of RAM (512 x 16 with byte enables)
assign dat[15:0] = (ram_sel[1] && ~we)?ram_1a_dat_o:{16{1'bZ}};
assign mem_dat[15:0] = (ram_sel[5] && ~mem_we)?ram_1b_dat_o:{16{1'bZ}};

ramb8_s16be_s16be 
  dp_ram_s16be_s16be_1 
  (
  .dat_o_s16a(ram_1a_dat_o),
  .adr_i_s16a(adr[9:1]),
  .dat_i_s16a(dat[15:0]),
  .be_i_a(2'b11),
  .rst_i_s16a(reset),
  .we_i_s16a(we && ram_sel[1]),
  .clk_i_s16a(sys_clk_0),
  .dat_o_s16b(ram_1b_dat_o),
  .adr_i_s16b(mem_adr[9:1]),
  .dat_i_s16b(mem_dat[15:0]),
  .be_i_b(mem_be[1:0]),
  .rst_i_s16b(reset),
  .we_i_s16b(mem_we & ram_sel[5]),
  .clk_i_s16b(sys_clk_0)
  );



// 2k bytes of RAM (512 x 32 with byte enables)
assign dat[31:0] = (ram_sel[2] && ~we)?ram_2a_dat_o:{32{1'bZ}};
assign mem_dat[31:0] = (ram_sel[6] && ~mem_we)?ram_2b_dat_o:{32{1'bZ}};

ramb16_s32be_s32be 
  dp_ram_s32be_s32be_1
  (
  .dat_o_s32a(ram_2a_dat_o),
  .adr_i_s32a(adr[10:2]),
  .dat_i_s32a(dat[31:0]),
  .be_i_a(4'b1111),
  .rst_i_s32a(reset),
  .we_i_s32a(we && ram_sel[2]),
  .clk_i_s32a(sys_clk_0),
  .dat_o_s32b(ram_2b_dat_o),
  .adr_i_s32b(mem_adr[10:2]),
  .dat_i_s32b(mem_dat[31:0]),
  .be_i_b(mem_be[3:0]),
  .rst_i_s32b(reset),
  .we_i_s32b(mem_we & ram_sel[6]),
  .clk_i_s32b(sys_clk_0)
  );



//--------------------------------------------------------------------------
// Module code
//--------------------------------------------------------------------------


endmodule

